/*
 * HYPEROID - a neato game
 *
 * Version: 1.1  Copyright (C) 1990,91 Hutchins Software
 *      This software is licenced under the GNU General Public Licence
 *      Please read the associated legal documentation
 * Author: Edward Hutchins
 * Internet: eah1@cec1.wustl.edu
 * USNail: c/o Edward Hutchins, 63 Ridgemoor Dr., Clayton, MO, 63105
 * Revisions:
 * 10/31/91 made game better/harder - Ed.
 *
 * Music: R.E.M./The Cure/Ministry/Front 242/The Smiths/New Order/Hendrix...
 * Beers: Bass Ale, Augsberger Dark
 */

/* Unix/Linux conversion by Russell Marks, 2000
 *
 * I think I'll add Rush to the Music list :-)
 */

/* Adapted for PalmOS by Russell Marks, 2002 */

#include <PalmOS.h>
#include <stdio.h>
#include <string.h>

#include "draw.h"
#include "main.h"
#include "misc.h"
#include "roidsupp.h"
#include "sincos.h"

#include "hyperoid.h"


#define PHYPEROID_CRID (('h'<<24)|('Y'<<16)|('p'<<8)|'d')
#define PHYPEROID_PREFS_VER	1


/* bitmaps for '0'..'9' */
static unsigned char numfontdat[50]=
  {
  0x44,0xAA,0xAA,0xAA,0x44,
  0x44,0xCC,0x44,0x44,0xEE,
  0xCC,0x22,0x44,0x88,0xEE,
  0xCC,0x22,0x44,0x22,0xCC,
  0x88,0xCC,0xEE,0x44,0x44,
  0xEE,0x88,0xCC,0x22,0xCC,
  0x66,0x88,0xCC,0xAA,0x44,
  0xEE,0x22,0x44,0x44,0x44,
  0x44,0xAA,0x44,0xAA,0x44,
  0x44,0xAA,0x66,0x22,0x44
  };

/* bitmap saying "PAUSE" */
static unsigned char pausedat[32]=
  {
  0x78,0xE6,0xCF,0x7C,0x6D,0xB6,0xD8,0x60,
  0x6D,0xB6,0xD8,0x60,0x6D,0xB6,0xCE,0x78,
  0x79,0xF6,0xC3,0x60,0x61,0xB6,0xC3,0x60,
  0x61,0xB6,0xC3,0x60,0x61,0xB3,0x9E,0x7C
  };


int safe_draw=1;
int paused=0;


/* vars needed for context */
int nDrawDelay,nLevel,nSafe,nBomb,nBadGuys;
long lScore,lLastLife,lHighScore;
int bRestart;

OBJ *npPlayer;
LIST FreeList,RoidList,ShotList,FlameList,SpinnerList;
LIST HunterList,HunterShotList,SwarmerList,LetterList;
OBJ Obj[MAX_OBJS];

static int restart_timer_count=0;


/* the duplication is horrible, but that's what you get for
 * retrofitting context-save onto an existing game. :-)
 */
struct game_save_tag
  {
  int nDrawDelay,nLevel,nSafe,nBomb,nBadGuys;
  long lScore,lLastLife;	/* lHighScore saved separately */
  int bRestart;
  
  OBJ *npPlayer;
  LIST lists[9];
  OBJ Obj[MAX_OBJS];

  int restart_timer_count;
  }
game_save;

/* struct for storing preferences, saved separately */
struct prefs_tag prefs;
int prefs_ok=0;		/* non-zero if contents of prefs struct is valid */



/* locals */

long dwSeed;
static RECT          rectShotClip;
static POINT         Player[] =
	{ {0, 0}, {160, 150}, {0, 250}, {96, 150}, {0, 0} };
static POINT         Spinner[] =
	{ {160, 150}, {224, 100}, {96, 100}, {32, 150}, {160, 150} };
static POINT         Swarmer[] =
	{ {0, 100}, {64, 100}, {128, 100}, {192, 100}, {0, 100} };
static POINT         Hunter[] =
{
	{160, 150}, {0, 250}, {192, 30}, {64, 30},
	{0, 250}, {96, 150}, {128, 150}, {160, 150}
};


/* KillBadGuy - kill off a badguy (made into a macro) */

#define KillBadGuy() \
	((--nBadGuys <= 0)?(SetRestart( RESTART_NEXTLEVEL ),TRUE):FALSE)



/* this oddity is needed to simulate the mapping the original used a
 * Windows call for. MAX_COORD divided by 80 is 64, so...
 */
#define convx(x)	(80+((x)>>6))
#define convy(y)	(80-((y)>>6))



void drawscore(long score,int left)
{
static unsigned char bmp[4*5];
char buf[16];
unsigned char *dst=bmp+(left?0:3),*src,*fnt;
int topnib=left;
int y;

memset(bmp,0,sizeof(bmp));
sprintf(buf+1,"%ld",score);
buf[1+8]=0;
*buf=0;
if(left)
  src=buf+1;
else
  src=buf+strlen(buf+1);

while(*src)
  {
  if(*src<'0' || *src>'9')
    {
    if(left) src++; else src--;
    continue;
    }
  fnt=numfontdat+(*src-'0')*5;
  for(y=0;y<5;y++,fnt++)
    dst[4*y]|=topnib?(*fnt&0xf0):(*fnt&0x0f);
  if(left) src++; else src--;
  topnib=!topnib;
  if(left && topnib) dst++;
  if(!left && !topnib) dst--;
  }

draw_bmp(left?0:128,0,32,5,bmp);
}


int IsKeyDown(int key)
{
switch(key)
  {
  case KEY_TAB:		return(0);
  case KEY_S:		return(keybmp&keyBitHard3);
  case KEY_LEFT:	return(keybmp&keyBitHard1);
  case KEY_RIGHT:	return(keybmp&keyBitHard2);
  case KEY_DOWN:	return(keybmp&keyBitPageDown);
  case KEY_UP:		return(keybmp&keyBitPageUp);
  case KEY_SPACE:	return(1);
  case KEY_ESC:		return(0);
  default:
    return(0);
  }
}


/* draw score and rest of status display */
void score_graphics(long score,long high,int lives,int bomb)
{
static unsigned char bmp[4*3];
unsigned char *dst;
int f,topnib;

drawscore(score,1);
drawscore(high,0);

memset(bmp,0,sizeof(bmp));
dst=bmp;
topnib=1;
for(f=0;f<lives && f<8;f++)
  {
  *dst  |=topnib?0x20:0x2;
  dst[4]|=topnib?0x70:0x7;
  dst[8]|=topnib?0x50:0x5;
  topnib=!topnib;
  if(topnib)
    dst++;
  }
draw_bmp(0,156,32,3,bmp);

memset(bmp,0,sizeof(bmp));
dst=bmp+3;
topnib=0;
for(f=0;f<bomb && f<8;f++)
  {
  *dst  |=topnib?0x40:0x4;
  dst[4]|=topnib?0xE0:0xE;
  dst[8]|=topnib?0x40:0x4;
  topnib=!topnib;
  if(!topnib)
    dst--;
  }
draw_bmp(128,156,32,3,bmp);
}


/* draw `paused' text over existing display buffer */
void drawpause(void)
{
draw_bmp(64,76,32,8,pausedat);
}



/* arand - pseudorandom number from 0 to x-1 (thanks antman!) */

int arand( int x )
{
	dwSeed = dwSeed * 0x343fd + 0x269ec3;
	return( (int)(((dwSeed >> 16) & 0x7fff) * x >> 15) );
}


/* AddHead - add an object to the head of a list */

void AddHead( LIST *npList, NODE *npNode )
{
	if (npList->npHead)
	{
		npNode->npNext = npList->npHead;
		npNode->npPrev = NULL;
		npList->npHead = (npList->npHead->npPrev = npNode);
	}
	else /* add to an empty list */
	{
		npList->npHead = npList->npTail = npNode;
		npNode->npNext = npNode->npPrev = NULL;
	}
}


/* RemHead - remove the first element in a list */

NODE *RemHead( LIST *npList )
{
	if (npList->npHead)
	{
		NODE *npNode = npList->npHead;
		if (npList->npTail != npNode)
		{
			npList->npHead = npNode->npNext;
			npNode->npNext->npPrev = NULL;
		}
		else npList->npHead = npList->npTail = NULL;
		return( npNode );
	}
	else return( NULL );
}


/* Remove - remove an arbitrary element from a list */

void Remove( LIST *npList, NODE *npNode )
{
	if (npNode->npPrev) npNode->npPrev->npNext = npNode->npNext;
	else npList->npHead = npNode->npNext;
	if (npNode->npNext) npNode->npNext->npPrev = npNode->npPrev;
	else npList->npTail = npNode->npPrev;
}


/* DrawObject - draw a single object */

void DrawObject( OBJ *npObj )
{
	int             nCnt;
	int             nDir = (npObj->nDir += npObj->nSpin);
	int             x = (npObj->Pos.x += npObj->Vel.x);
	int             y = (npObj->Pos.y += npObj->Vel.y);
	POINT           Pts[MAX_PTS];

	if (x < -CLIP_COORD) npObj->Pos.x = x = CLIP_COORD;
	else if (x > CLIP_COORD) npObj->Pos.x = x = -CLIP_COORD;
	if (y < -CLIP_COORD) npObj->Pos.y = y = CLIP_COORD;
	else if (y > CLIP_COORD) npObj->Pos.y = y = -CLIP_COORD;

	for (nCnt = npObj->byPts - 1; nCnt >= 0; --nCnt)
	{
		int wDeg = DEG( npObj->Pts[nCnt].x + nDir );
		int nLen = npObj->Pts[nCnt].y;
		Pts[nCnt].x = x + MULDEG( nLen, nCos[wDeg] );
		Pts[nCnt].y = y + MULDEG( nLen, nSin[wDeg] );
	}

	if (npObj->byPts > 1)
	{
        	int n=npObj->byPts;
        
		if (npObj->nCount > 0 && n>=2)
                {
                POINT *pts=&(Pts[0]);
                int f,x,y,ox=convx(pts->x),oy=convy(pts->y);

                pts++;
                for(f=1;f<n;f++,pts++)
                  {
                  x=convx(pts->x); y=convy(pts->y);
                  draw_line(ox,oy,x,y);
                  ox=x; oy=y;
                  }
                }
	}
	else /* just a point */
	{
		if (npObj->nCount > 0)
                  draw_pixel( convx(Pts[0].x), convy(Pts[0].y) );
	}
}


/* SetRestart - set the restart timer */

void SetRestart( RESTART_MODE Restart )
{
	POINT           Pt;
	char            szBuff[32];

	if (bRestart) return;
        restart_timer_count=RESTART_DELAY_FRAMES;
	bRestart = TRUE;

	Pt.x = Pt.y = 0;
	switch (Restart)
	{
	case RESTART_GAME:
		SpinLetters( "GAME OVER", Pt, Pt, 400 );
		break;
	case RESTART_LEVEL:
		PrintLetters( "GET READY", Pt, Pt, 300 );
		break;
	case RESTART_NEXTLEVEL:
		sprintf( szBuff, "LEVEL %d", nLevel + 1 );
		PrintLetters( szBuff, Pt, Pt, 300 );
		break;
	}
}


/* PrintPlayerMessage - show the player a status message */

void PrintPlayerMessage( char * npszText )
{
	POINT Pos, Vel;

	Pos = npPlayer->Pos;
	Pos.y -= 400;
	Vel.x = 0;
	Vel.y = -50;
	PrintLetters( npszText, Pos, Vel, 150 );
}


/* AddExtraLife - give the player another life */

void AddExtraLife( void )
{
	PrintPlayerMessage( "EXTRA LIFE" );
	++npPlayer->nCount;
}


/* Hit - something hit an object, do fireworks */

void Hit( OBJ *npObj )
{
	int             nCnt;

	for (nCnt = 0; nCnt < 6; ++nCnt)
	{
		OBJ *npFlame = RemHeadObj( &FreeList );
		if (!npFlame) return;
		npFlame->Pos.x = npObj->Pos.x;
		npFlame->Pos.y = npObj->Pos.y;
		npFlame->Vel.x = npObj->Vel.x;
		npFlame->Vel.y = npObj->Vel.y;
		npFlame->nDir = npObj->nDir + (nCnt * DEGREE_SIZE) / 6;
		npFlame->nSpin = 0;
		npFlame->nCount = 10 + arand( 8 );
		npFlame->byPts = 1;
		npFlame->Pts[0].x = npFlame->Pts[0].y = 0;
		ACCEL( npFlame, npFlame->nDir, 50 - npFlame->nCount );
		AddHeadObj( &FlameList, npFlame );
	}
}


/* Explode - explode an object */

void Explode( OBJ *npObj )
{
	int             nCnt, nSize = npObj->byPts;

	DrawObject( npObj );
	for (nCnt = 0; nCnt < nSize; ++nCnt)
	{
		OBJ *npFlame;
		if (arand( 2 )) continue;
		if (!(npFlame = RemHeadObj( &FreeList ))) return;
		npFlame->Pos.x = npObj->Pos.x;
		npFlame->Pos.y = npObj->Pos.y;
		npFlame->Vel.x = npObj->Vel.x;
		npFlame->Vel.y = npObj->Vel.y;
		npFlame->nDir = npObj->nDir + nCnt * DEGREE_SIZE / nSize + arand( 32 );
		npFlame->nSpin = arand( 31 ) - 15;
		npFlame->nCount = 25 + arand( 16 );
		npFlame->byPts = 2;
		npFlame->Pts[0] = npObj->Pts[nCnt];
		if (nCnt == nSize - 1) npFlame->Pts[1] = npObj->Pts[0];
		else npFlame->Pts[1] = npObj->Pts[nCnt + 1];
		ACCEL( npFlame, npFlame->nDir, 60 - npFlame->nCount );
		AddHeadObj( &FlameList, npFlame );
	}
	Hit( npObj );
}


/* HitPlayer - blow up the player */

int HitPlayer( OBJ *npObj )
{
	POINT           Vel;
	int             nMass, nSpin;

	if (nSafe || (npPlayer->nCount <= 0)) return( FALSE );

	/* rumble and shake both objects */
	nMass = npPlayer->nMass + npObj->nMass;

	nSpin = npPlayer->nSpin + npObj->nSpin;
	npObj->nSpin -= MulDiv( nSpin, npPlayer->nMass, nMass );
	npPlayer->nSpin -= MulDiv( nSpin, npObj->nMass, nMass );

	Vel.x = npPlayer->Vel.x - npObj->Vel.x;
	Vel.y = npPlayer->Vel.y - npObj->Vel.y;
	npObj->Vel.x += MulDiv( Vel.x, npPlayer->nMass, nMass );
	npObj->Vel.y += MulDiv( Vel.y, npPlayer->nMass, nMass );
	npPlayer->Vel.x -= MulDiv( Vel.x, npObj->nMass, nMass );
	npPlayer->Vel.y -= MulDiv( Vel.y, npObj->nMass, nMass );

	if (--npPlayer->nCount)
	{
		Hit( npPlayer );
		return( TRUE );
	}

	/* final death */
	Explode( npPlayer );
	SetRestart( RESTART_GAME );
	return( FALSE );
}


/* CreateLetter - make a new letter object */

OBJ *CreateLetter( int cLetter, int nSize )
{
	OBJ *npLtr;
	int nCnt;
	char *npDesc;

	if (cLetter >= '0' && cLetter <= '9') npDesc = NumberDesc[cLetter - '0'];
	else if (cLetter >= 'A' && cLetter <= 'Z') npDesc = LetterDesc[cLetter - 'A'];
	else if (cLetter >= 'a' && cLetter <= 'z') npDesc = LetterDesc[cLetter - 'a'];
	else if (cLetter == '.') npDesc = "l";
	else return( NULL );

	if ((npLtr = RemHeadObj( &FreeList )))
	{
		npLtr->nMass = 1;
		npLtr->nDir = 0;
		npLtr->nSpin = 0;
		npLtr->nCount = 40;
		npLtr->byPts = (BYTE)(nCnt = strlen( npDesc ));
		while (nCnt--)
		{
			npLtr->Pts[nCnt] = LetterPart[npDesc[nCnt] - 'a'];
			npLtr->Pts[nCnt].y = MulDiv( (long)npLtr->Pts[nCnt].y, nSize, LETTER_MAX );
		}
		AddHeadObj( &LetterList, npLtr );
	}
	return( npLtr );
}


/* DrawLetters - draw letters and such */

void DrawLetters( void )
{
	OBJ *npLtr, *npNext;

	for (npLtr = HeadObj( &LetterList ); npLtr; npLtr = npNext)
	{
		npNext = NextObj( npLtr );
		if (--npLtr->nCount == 0)
		{
			RemoveObj( &LetterList, npLtr );
			AddHeadObj( &FreeList, npLtr );
		}
		DrawObject( npLtr );
	}
}


/* DrawHunterShots - process and draw the hunter shot list */

void DrawHunterShots( void )
{
	OBJ *npShot, *npNext;

	for (npShot = HeadObj( &HunterShotList ); npShot; npShot = npNext)
	{
		RECT            rect;

		npNext = NextObj( npShot );

		MKRECT( &rect, npShot->Pos, 200 );

		if (PTINRECT( &rect, npPlayer->Pos ))
		{
			HitPlayer( npShot );
			npShot->nCount = 1;
		}
		if (--npShot->nCount == 0)
		{
			RemoveObj( &HunterShotList, npShot );
			AddHeadObj( &FreeList, npShot );
		}
		DrawObject( npShot );
	}
}


/* FireHunterShot - fire a hunter bullet */

void FireHunterShot( OBJ *npHunt )
{
	OBJ *          npShot;

	if ((npShot = RemHeadObj( &FreeList )))
	{
		npShot->Pos.x = npHunt->Pos.x;
		npShot->Pos.y = npHunt->Pos.y;
		npShot->Vel.x = npHunt->Vel.x;
		npShot->Vel.y = npHunt->Vel.y;
		npShot->nMass = 8;
		npShot->nDir = npHunt->nDir + arand( 5 ) - 2;
		npShot->nSpin = (arand( 2 ) ? 10 : -10);
		npShot->nCount = 16 + arand( 8 );
		npShot->byPts = 2;
		npShot->Pts[0].x = 128;
		npShot->Pts[0].y = 50;
		npShot->Pts[1].x = 0;
		npShot->Pts[1].y = 50;
		ACCEL( npShot, npShot->nDir, 200 + npShot->nCount );
		AddHeadObj( &HunterShotList, npShot );
	}
}


/* CreateHunter - make a new hunter */

void CreateHunter( void )
{
	OBJ *          npHunt;
	int             nCnt;

	if ((npHunt = RemHeadObj( &FreeList )))
	{
		npHunt->Pos.x = arand( CLIP_COORD * 2 ) - CLIP_COORD;
		npHunt->Pos.y = -CLIP_COORD;
		npHunt->Vel.x = npHunt->Vel.y = 0;
		npHunt->nMass = 256;
		npHunt->nDir = arand( DEGREE_SIZE );
		npHunt->nSpin = 0;
		npHunt->nCount = 1 + arand( (nLevel>2)?2:nLevel );
		npHunt->nDelay = 2 + arand( 10 );
		npHunt->byPts = DIM(Hunter);
		for (nCnt = 0; nCnt < DIM(Hunter); ++nCnt)
			npHunt->Pts[nCnt] = Hunter[nCnt];
		ACCEL( npHunt, npHunt->nDir, 30 + nLevel * 2 );
		AddHeadObj( &HunterList, npHunt );
		++nBadGuys;
	}
}


/* DrawHunters - process and draw the hunter list */

void DrawHunters( void )
{
	OBJ *npHunt, *npNext;
	static int       nNextHunter = 200;

	if (nBadGuys && (--nNextHunter < 0))
	{
		CreateHunter();
		nNextHunter = 1000 + arand( 1000 ) - nLevel * 8;
	}

	for (npHunt = HeadObj( &HunterList ); npHunt; npHunt = npNext)
	{
		OBJ *          npShot;
		RECT            rect;

		npNext = NextObj( npHunt );

		MKRECT( &rect, npHunt->Pos, 200 );

		if (PTINRECT( &rect, npPlayer->Pos ))
		{
			HitPlayer( npHunt );
			--npHunt->nCount;
			if (npHunt->nCount < 1)
			{
				KillBadGuy();
				Explode( npHunt );
				RemoveObj( &HunterList, npHunt );
				AddHeadObj( &FreeList, npHunt );
			}
		}
		else if (INTRECT(&rect, &rectShotClip))
		{
			for (npShot = HeadObj( &ShotList ); npShot; npShot = NextObj( npShot ))
			{
				if (!PTINRECT( &rect, npShot->Pos )) continue;
				npShot->nCount = 1;
				lScore += npHunt->nCount * 1000;
				if (--npHunt->nCount < 1)
				{
					KillBadGuy();
					Explode( npHunt );
					RemoveObj( &HunterList, npHunt );
					AddHeadObj( &FreeList, npHunt );
				}
				else
				{
					Hit( npHunt );
				}
				break;
			}
		}
		ACCEL( npHunt, npHunt->nDir, 8 );
		npHunt->Vel.x -= npHunt->Vel.x / 16;
		npHunt->Vel.y -= npHunt->Vel.y / 16;
		if (--npHunt->nDelay <= 0)
		{
			npHunt->nDelay = arand( 10 );
			npHunt->nSpin = arand( 11 ) - 5;
			FireHunterShot( npHunt );
		}
		DrawObject( npHunt );
	}
}


/* CreateSwarmer - make a new swarmer */

void CreateSwarmer( POINT Pos, int nDir, int nCount )
{
	OBJ *          npSwarm;
	int             nCnt;

	if ((npSwarm = RemHeadObj( &FreeList )))
	{
		npSwarm->Pos = Pos;
		npSwarm->Vel.x = npSwarm->Vel.y = 0;
		npSwarm->nDir = nDir;
		npSwarm->nSpin = arand( 31 ) - 15;
		npSwarm->nCount = nCount;
		npSwarm->nDelay = 64 + arand( 64 );
		npSwarm->nMass = 32;
		npSwarm->byPts = DIM(Swarmer);
		for (nCnt = 0; nCnt < DIM(Swarmer); ++nCnt)
		{
			npSwarm->Pts[nCnt] = Swarmer[nCnt];
			npSwarm->Pts[nCnt].y += nCount * 10;
		}
		ACCEL( npSwarm, npSwarm->nDir, 30 + nLevel * 2 );
		AddHeadObj( &SwarmerList, npSwarm );
		++nBadGuys;
	}
}


/* DrawSwarmers - process and draw the swarmer list */

void DrawSwarmers( void )
{
	OBJ *npSwarm, *npNext;
	static int       nNextSwarmer = 1000;

	if (nBadGuys && (--nNextSwarmer < 0))
	{
		POINT Pos;
		Pos.x = arand( CLIP_COORD * 2 ) - CLIP_COORD;
		Pos.y = -CLIP_COORD;
		CreateSwarmer( Pos, arand( DEGREE_SIZE ), 8 + nLevel * 2 );
		nNextSwarmer = 1000 + arand( 500 ) - nLevel * 4;
	}

	for (npSwarm = HeadObj( &SwarmerList ); npSwarm; npSwarm = npNext)
	{
		OBJ *          npShot;
		RECT            rect;

		npNext = NextObj( npSwarm );

		MKRECT( &rect, npSwarm->Pos, 150 + npSwarm->nCount * 10 );

		if (PTINRECT( &rect, npPlayer->Pos ))
		{
			HitPlayer( npSwarm );
			npSwarm->nCount = 0;
		}
		else if (INTRECT(&rect, &rectShotClip))
		{
			for (npShot = HeadObj( &ShotList ); npShot; npShot = NextObj( npShot ))
			{
				if (!PTINRECT( &rect, npShot->Pos )) continue;
				npShot->nCount = 1;
				lScore += npSwarm->nCount * 25;
				npSwarm->nCount = 0;
				break;
			}
		}
		if (npSwarm->nCount <= 0)
		{
			KillBadGuy();
			Explode( npSwarm );
			RemoveObj( &SwarmerList, npSwarm );
			AddHeadObj( &FreeList, npSwarm );
		}
		else
		{
			if ((npSwarm->nCount > 1) && (--npSwarm->nDelay <= 0))
			{
				int nDir = arand( DEGREE_SIZE );
				int nCount = npSwarm->nCount / 2;
				CreateSwarmer( npSwarm->Pos, nDir, nCount );
				nCount = npSwarm->nCount - nCount;
				CreateSwarmer( npSwarm->Pos, nDir + 128, nCount );
				npSwarm->nCount = 0;
			}
			DrawObject( npSwarm );
		}
	}
}


/* CreateSpinner - make a new spinner */

void CreateSpinner( void )
{
	OBJ *          npSpin;
	int             nCnt;

	if ((npSpin = RemHeadObj( &FreeList )))
	{
		npSpin->Pos.x = arand( CLIP_COORD * 2 ) - CLIP_COORD;
		npSpin->Pos.y = -CLIP_COORD;
		npSpin->Vel.x = npSpin->Vel.y = 0;
		npSpin->nDir = arand( DEGREE_SIZE );
		npSpin->nSpin = -12;
		npSpin->nCount = 1 + arand( (nLevel>2)?2:nLevel );
		npSpin->nMass = 64 + npSpin->nCount * 32;
		npSpin->byPts = DIM(Spinner);
		for (nCnt = 0; nCnt < DIM(Spinner); ++nCnt)
			npSpin->Pts[nCnt] = Spinner[nCnt];
		ACCEL( npSpin, npSpin->nDir, 30 + nLevel * 2 );
		AddHeadObj( &SpinnerList, npSpin );
		++nBadGuys;
	}
}


/* DrawSpinners - process and draw the spinner list */

void DrawSpinners( void )
{
	OBJ *npSpin, *npNext;
	static int       nNextSpinner = 1000;

	if (nBadGuys && (--nNextSpinner < 0))
	{
		CreateSpinner();
		nNextSpinner = 100 + arand( 900 ) - nLevel * 2;
	}

	for (npSpin = HeadObj( &SpinnerList ); npSpin; npSpin = npNext)
	{
		OBJ *          npShot;
		int             nDelta;
		RECT            rect;

		npNext = NextObj( npSpin );

		MKRECT( &rect, npSpin->Pos, 150 );

		if (PTINRECT( &rect, npPlayer->Pos ))
		{
			HitPlayer( npSpin );
			--npSpin->nCount;
			if (npSpin->nCount < 1)
			{
				KillBadGuy();
				Explode( npSpin );
				RemoveObj( &SpinnerList, npSpin );
				AddHeadObj( &FreeList, npSpin );
			}
		}
		else if (INTRECT(&rect, &rectShotClip))
		{
			for (npShot = HeadObj( &ShotList ); npShot; npShot = NextObj( npShot ))
			{
				if (!PTINRECT( &rect, npShot->Pos )) continue;
				npShot->nCount = 1;
				lScore += npSpin->nCount * 500;
				if (--npSpin->nCount < 1)
				{
					KillBadGuy();
					Explode( npSpin );
					RemoveObj( &SpinnerList, npSpin );
					AddHeadObj( &FreeList, npSpin );
				}
				else
				{
					Hit( npSpin );
				}
				break;
			}
		}
		nDelta = npPlayer->Pos.x - npSpin->Pos.x;
		while (nDelta < -16 || nDelta > 16) nDelta /= 2;
		npSpin->Vel.x += nDelta - npSpin->Vel.x / 16;
		nDelta = npPlayer->Pos.y - npSpin->Pos.y;
		while (nDelta < -16 || nDelta > 16) nDelta /= 2;
		npSpin->Vel.y += nDelta - npSpin->Vel.y / 16;
		DrawObject( npSpin );
	}
}


/* CreateRoid - make a new asteroid */

void CreateRoid( POINT Pos, POINT Vel, int nSides,
					 int nDir, int nSpeed, int nSpin )
{
	OBJ *          npRoid;
	int             nCnt;

	if ((npRoid = RemHeadObj( &FreeList )))
	{
		npRoid->Pos = Pos;
		npRoid->Vel = Vel;
		npRoid->nMass = nSides * 128;
		npRoid->nDir = nDir;
		npRoid->nSpin = nSpin + arand( 11 ) - 5;
		npRoid->nCount = nSides * 100;
		npRoid->byPts = (BYTE)(nSides + 1);
		for (nCnt = 0; nCnt < nSides; ++nCnt)
		{
			npRoid->Pts[nCnt].x = nCnt * DEGREE_SIZE / nSides + arand( 30 );
			npRoid->Pts[nCnt].y = (nSides - 1) * 100 + 20 + arand( 80 );
		}
		npRoid->Pts[nSides] = npRoid->Pts[0];
		ACCEL( npRoid, nDir, nSpeed );
		AddHeadObj( &RoidList, npRoid );
		++nBadGuys;
	}
}


/* BreakRoid - break up an asteroid */

void BreakRoid( OBJ *npRoid, OBJ *npShot )
{
	int             nCnt, nNew;

	lScore += npRoid->nCount;
	if (npShot) npShot->nCount = 1;
	switch (npRoid->byPts)
	{
	case 8:
		nNew = 2 + arand( 3 );
		break;
	case 7:
		nNew = 1 + arand( 3 );
		break;
	case 6:
		nNew = 1 + arand( 2 );
		break;
	case 5:
		nNew = arand( 2 );
		break;
	default:
		nNew = 0;
		break;
	}
	if (nNew == 1)		/* don't explode outward */
	{
		POINT Pt = npRoid->Pos;
		Pt.x += arand( 301 ) - 150; Pt.y += arand( 301 ) - 150;
		CreateRoid( Pt, npRoid->Vel, npRoid->byPts - (nNew + 1),
			npShot?(npShot->nDir):npRoid->nDir,
			8, npRoid->nSpin );
	}
	else if (nNew > 0)
	{
		int nSpeed = npRoid->nSpin * npRoid->nSpin * nNew + 16;
		for (nCnt = 0; nCnt < nNew; ++nCnt)
		{
			POINT Pt = npRoid->Pos;
			Pt.x += arand( 601 ) - 300; Pt.y += arand( 601 ) - 300;
			CreateRoid( Pt, npRoid->Vel, npRoid->byPts - (nNew + 1),
						npRoid->nDir + nCnt * DEGREE_SIZE / nNew + arand( 32 ),
						nSpeed + arand( nLevel * 4 ),
						npRoid->nSpin / 2 );
		}
	}
	KillBadGuy();
	npRoid->nCount = 0;
	if (nNew)
	{
		Hit( npRoid );
		DrawObject( npRoid );
	}
	else
        {
        	Explode( npRoid );
	}
	RemoveObj( &RoidList, npRoid );
	AddHeadObj( &FreeList, npRoid );
}


/* DrawRoids - process and draw the asteroid list */

void DrawRoids( void )
{
	OBJ *npRoid, *npNext;

	for (npRoid = HeadObj( &RoidList ); npRoid; npRoid = npNext)
	{
		int             nSize = npRoid->nCount;
		OBJ *          npShot;
		RECT            rect;

		npNext = NextObj( npRoid );

		DrawObject( npRoid );

		MKRECT( &rect, npRoid->Pos, nSize );

		if (PTINRECT( &rect, npPlayer->Pos ) && HitPlayer( npRoid ))
		{
			npPlayer->nCount = -npPlayer->nCount;
			Explode( npPlayer );
			BreakRoid( npRoid, NULL );
			if (nBadGuys) SetRestart( RESTART_LEVEL );
			else SetRestart( RESTART_NEXTLEVEL );
		}
		else if (INTRECT(&rect, &rectShotClip))
		{
			for (npShot = HeadObj( &ShotList ); npShot; npShot = NextObj( npShot ))
			{
				if (!PTINRECT( &rect, npShot->Pos )) continue;
				BreakRoid( npRoid, npShot );
				break;
			}
		}
	}
}


/* DrawShots - process and draw the player shot list */

void DrawShots( void )
{
	OBJ *npShot, *npNext;

	if ((npShot = HeadObj( &ShotList )))
	{
		rectShotClip.left = rectShotClip.right = npShot->Pos.x;
		rectShotClip.top = rectShotClip.bottom = npShot->Pos.y;
		while (npShot)
		{
			npNext = NextObj( npShot );
			if (--npShot->nCount == 0)
			{
				RemoveObj( &ShotList, npShot );
				AddHeadObj( &FreeList, npShot );
			}
			DrawObject( npShot );
			if (npShot->Pos.x < rectShotClip.left) rectShotClip.left = npShot->Pos.x;
			else if (npShot->Pos.x > rectShotClip.right) rectShotClip.right = npShot->Pos.x;
			if (npShot->Pos.y < rectShotClip.top) rectShotClip.top = npShot->Pos.y;
			else if (npShot->Pos.y > rectShotClip.bottom) rectShotClip.bottom = npShot->Pos.y;
			npShot = npNext;
		}
	}
	else rectShotClip.left = rectShotClip.right = rectShotClip.top = rectShotClip.bottom = 32767;
}


/* DrawFlames - process and draw the flame list */

void DrawFlames( void )
{
	OBJ *npFlame, *npNext;

	for (npFlame = HeadObj( &FlameList ); npFlame; npFlame = npNext)
	{
		npNext = NextObj( npFlame );
		if (--npFlame->nCount == 0)
		{
			RemoveObj( &FlameList, npFlame );
			AddHeadObj( &FreeList, npFlame );
		}
		DrawObject( npFlame );
	}
}


/* FireShot - fire a bullet */

void FireShot( void )
{
	OBJ *          npShot;

	if ((npShot = RemHeadObj( &FreeList )))
	{
		npShot->Pos.x = npPlayer->Pos.x;
		npShot->Pos.y = npPlayer->Pos.y;
		npShot->Vel.x = npPlayer->Vel.x;
		npShot->Vel.y = npPlayer->Vel.y;
		npShot->nMass = 8;
		npShot->nDir = npPlayer->nDir + arand( 5 ) - 2;
		npShot->nSpin = 0;
		npShot->nCount = 16 + arand( 8 );
		npShot->byPts = 2;
		npShot->Pts[0].x = 128;
		npShot->Pts[0].y = 50;
		npShot->Pts[1].x = 0;
		npShot->Pts[1].y = 50;
		ACCEL( npShot, npShot->nDir, 200 + npShot->nCount );
		AddHeadObj( &ShotList, npShot );
	}
}


/* AccelPlayer - move the player forward */

void AccelPlayer( int nDir, int nAccel )
{
	OBJ *          npFlame;

	nDir += npPlayer->nDir;
	if (nAccel) ACCEL( npPlayer, nDir, nAccel );
	if ((npFlame = RemHeadObj( &FreeList )))
	{
		npFlame->Pos.x = npPlayer->Pos.x;
		npFlame->Pos.y = npPlayer->Pos.y;
		npFlame->Vel.x = npPlayer->Vel.x;
		npFlame->Vel.y = npPlayer->Vel.y;
		npFlame->nDir = nDir + 100 + arand( 57 );
		npFlame->nSpin = 0;
		npFlame->nCount = nAccel + arand( 7 );
		npFlame->byPts = 1;
		npFlame->Pts[0].x = npFlame->Pts[0].y = 0;
		ACCEL( npFlame, npFlame->nDir, 50 + arand( 10 ) );
		AddHeadObj( &FlameList, npFlame );
	}
}


/* HitList - Hit() a list of things */

void HitList( LIST *npList )
{
	OBJ *          npObj;

	for (npObj = HeadObj( npList ); npObj; npObj = NextObj( npObj ))
		if (npObj->nCount) Hit( npObj );
}


/* ExplodeBadguys - explode a list of badguys */

void ExplodeBadguys( LIST *npList )
{
	OBJ *          npObj;

	while ((npObj = HeadObj( npList )))
	{
		KillBadGuy();
		npObj->nCount = 0;
		Explode( npObj );
		RemoveObj( npList, npObj );
		AddHeadObj( &FreeList, npObj );
	}
}


/* DrawPlayer - process and draw the player */

void DrawPlayer( void )
{
	static int       nBombing = 0;
	static int       nShotDelay = 0;

	if (npPlayer->nCount <= 0) return;

	if (nSafe > 0) --nSafe;

	if (nBombing > 0)
	{
		if (--nBombing == 0)
		{
			ExplodeBadguys( &SpinnerList );
			ExplodeBadguys( &SwarmerList );
			ExplodeBadguys( &HunterList );
		}
		else
		{
			HitList( &SpinnerList );
			HitList( &SwarmerList );
			HitList( &HunterList );
		}
	}
	else if (nBomb && IsKeyDown( KEY_S )) --nBomb, nBombing = 5;

	if (IsKeyDown( KEY_LEFT )) npPlayer->nSpin += 8;
	if (IsKeyDown( KEY_RIGHT )) npPlayer->nSpin -= 8;
	if (IsKeyDown( KEY_DOWN )) AccelPlayer( 0, 12 );
	if (IsKeyDown( KEY_UP )) AccelPlayer( 128, 12 );
	if (nShotDelay) --nShotDelay;
	else if (IsKeyDown( KEY_SPACE )) FireShot(), nShotDelay = 2;
        if(!nSafe || safe_draw)
	  DrawObject( npPlayer );
	npPlayer->nSpin /= 2;
}



/* DrawObjects - transform and redraw everything in the system */

void DrawObjects( void )
{
	/* move and draw things (I don't think the order is important...) */
	DrawPlayer();
	DrawFlames();
	DrawShots();
	DrawRoids();
	DrawSpinners();
	DrawSwarmers();
	DrawHunters();
	DrawHunterShots();
	DrawLetters();
	/* (...but I'm not changing it!!! :-) */
}


/* CheckScore - show the score and such stuff */

void CheckScore( void )
{
	int nLives;

	if (lScore - lLastLife > EXTRA_LIFE)
	{
		AddExtraLife();
		lLastLife = lScore;
	}

	/* apparently, -ve player lives means we're starting a new
	 * life soon (ouch). -rjm
	 */
	nLives=((npPlayer->nCount > 0) ? npPlayer->nCount : -npPlayer->nCount);

	/* actually do the score/lives/etc-drawing */
	score_graphics(lScore,lHighScore,nLives,nBomb);
}


/* NewGame - start a new game */

void NewGame( void )
{
	npPlayer->nCount = 0;
	Explode( npPlayer );
	SetRestart( RESTART_GAME );
	ExplodeBadguys( &RoidList );
	ExplodeBadguys( &SpinnerList );
	ExplodeBadguys( &SwarmerList );
	ExplodeBadguys( &HunterList );
}


/* RestartHyperoid - set up a game! */

void RestartHyperoid( void )
{
	if (npPlayer->nCount == 0)
	{
		POINT Pos, Vel;
		Pos.x = 0;
		Pos.y = -CLIP_COORD / 2;
		Vel.x = 0;
		Vel.y = 150;
		PrintLetters( "HYPEROID", Pos, Vel, 800 );
		npPlayer->nCount = 3;
		if (lHighScore < lScore) lHighScore = lScore;
		lLastLife = lScore = 0;
		nLevel = 0;
		nBomb = 3;
	}
	else if (npPlayer->nCount < 0)
	{
		/* cheesy way of restarting after a major collision */
		npPlayer->nCount = -npPlayer->nCount;
		nBomb = 3;
	}

	npPlayer->Pos.x = npPlayer->Pos.y = 0;
	npPlayer->Vel.x = npPlayer->Vel.y = 0;
	npPlayer->nDir = 64;
	npPlayer->nSpin = 0;
	nSafe = 30;

	if (ShotList.npHead)
	{
		OBJ *npShot;
		for (npShot = HeadObj( &ShotList ); npShot; npShot = NextObj( npShot ))
			npShot->nCount = 1;
	}

	/* reseed the asteroid field */
	if (nBadGuys == 0)
	{
		int nCnt;
		++nLevel;
		for (nCnt = 5 + nLevel; nCnt; --nCnt)
		{
			POINT Pos, Vel;
			Pos.x = arand( MAX_COORD * 2 ) - MAX_COORD;
			Pos.y = arand( MAX_COORD * 2 ) - MAX_COORD;
			Vel.x = Vel.y = 0;
			CreateRoid( Pos, Vel, 6 + arand( 2 ),
						arand( DEGREE_MAX ), 30 + arand( nLevel * 8 ), 0 );
		}
	}
}



/* InitHyperoid - initialize everything */

void InitHyperoid( void )
{
	int nCnt;

	/* seed the randomizer */
	dwSeed = TimGetTicks();

	/* allocate all objects as free */
	for (nCnt = 0; nCnt < MAX_OBJS; ++nCnt)
		AddHeadObj( &FreeList, &(Obj[nCnt]) );

	/* set up the player */
	npPlayer = RemHeadObj( &FreeList );
	npPlayer->byPts = DIM(Player);
	npPlayer->nMass = 256;
	for (nCnt = 0; nCnt < DIM(Player); ++nCnt)
		npPlayer->Pts[nCnt] = Player[nCnt];
}


/* the +1 is to differentiate from NULL pointer, which is left alone */
#define ptr_to_ofs(x)	((x)?((void *)(((UInt32)(x))+1-(UInt32)Obj)):NULL)
#define ofs_to_ptr(x)	((x)?((void *)(((UInt32)(x))-1+(UInt32)Obj)):NULL)


void save_vars(void)
{
int f;

game_save.nDrawDelay=nDrawDelay;
game_save.nLevel=nLevel;
game_save.nSafe=nSafe;
game_save.nBomb=nBomb;
game_save.nBadGuys=nBadGuys;
game_save.lScore=lScore;
game_save.lLastLife=lLastLife;
game_save.bRestart=bRestart;
game_save.npPlayer=ptr_to_ofs(npPlayer);

game_save.lists[0]=FreeList;
game_save.lists[1]=RoidList;
game_save.lists[2]=ShotList;
game_save.lists[3]=FlameList;
game_save.lists[4]=SpinnerList;
game_save.lists[5]=HunterList;
game_save.lists[6]=HunterShotList;
game_save.lists[7]=SwarmerList;
game_save.lists[8]=LetterList;

memcpy(game_save.Obj,Obj,sizeof(Obj));

game_save.restart_timer_count=restart_timer_count;

/* convert lists */
for(f=0;f<9;f++)
  {
  game_save.lists[f].npHead=ptr_to_ofs(game_save.lists[f].npHead);
  game_save.lists[f].npTail=ptr_to_ofs(game_save.lists[f].npTail);
  }

/* convert Obj[] nodes */
for(f=0;f<MAX_OBJS;f++)
  {
  game_save.Obj[f].Link.npNext=ptr_to_ofs(game_save.Obj[f].Link.npNext);
  game_save.Obj[f].Link.npPrev=ptr_to_ofs(game_save.Obj[f].Link.npPrev);
  }

PrefSetAppPreferences(PHYPEROID_CRID,0,PHYPEROID_PREFS_VER,
                      &game_save,sizeof(game_save),false);
PrefSetAppPreferences(PHYPEROID_CRID,0,PHYPEROID_PREFS_VER,
                      &lHighScore,sizeof(lHighScore),true);
PrefSetAppPreferences(PHYPEROID_CRID,1,PHYPEROID_PREFS_VER,
                      &prefs,sizeof(prefs),true);
}


/* returns non-zero if we successfully read context, and thus
 * should pause on startup.
 */
int restore_vars(void)
{
UInt16 siz;
Int16 ver;
int f;

/* do prefs first */
siz=sizeof(prefs);
ver=PrefGetAppPreferences(PHYPEROID_CRID,1,&prefs,&siz,true);
if(ver!=PHYPEROID_PREFS_VER || siz!=sizeof(prefs))
  {
  /* if we couldn't restore them, set up defaults */
  prefs.invert=0;
  prefs.fg=0;
  prefs.bg=7;
  memset(prefs.reserved,0,sizeof(prefs.reserved));
  }
prefs_ok=1;

/* restore high score */
siz=sizeof(lHighScore);
ver=PrefGetAppPreferences(PHYPEROID_CRID,0,&lHighScore,&siz,true);
if(ver!=PHYPEROID_PREFS_VER || siz!=sizeof(lHighScore))
  return(0);

/* if the high score was ok, try game context */
siz=sizeof(game_save);
ver=PrefGetAppPreferences(PHYPEROID_CRID,0,&game_save,&siz,false);
if(ver!=PHYPEROID_PREFS_VER || siz!=sizeof(game_save))
  return(0);

/* convert lists */
for(f=0;f<9;f++)
  {
  game_save.lists[f].npHead=ofs_to_ptr(game_save.lists[f].npHead);
  game_save.lists[f].npTail=ofs_to_ptr(game_save.lists[f].npTail);
  }

/* convert Obj[] nodes */
for(f=0;f<MAX_OBJS;f++)
  {
  game_save.Obj[f].Link.npNext=ofs_to_ptr(game_save.Obj[f].Link.npNext);
  game_save.Obj[f].Link.npPrev=ofs_to_ptr(game_save.Obj[f].Link.npPrev);
  }

nDrawDelay=game_save.nDrawDelay;
nLevel=game_save.nLevel;
nSafe=game_save.nSafe;
nBomb=game_save.nBomb;
nBadGuys=game_save.nBadGuys;
lScore=game_save.lScore;
lLastLife=game_save.lLastLife;
bRestart=game_save.bRestart;
npPlayer=ofs_to_ptr(game_save.npPlayer);

FreeList=game_save.lists[0];
RoidList=game_save.lists[1];
ShotList=game_save.lists[2];
FlameList=game_save.lists[3];
SpinnerList=game_save.lists[4];
HunterList=game_save.lists[5];
HunterShotList=game_save.lists[6];
SwarmerList=game_save.lists[7];
LetterList=game_save.lists[8];

memcpy(Obj,game_save.Obj,sizeof(Obj));

restart_timer_count=game_save.restart_timer_count;

return(1);
}


/* wait for keys to be released */
int event_loop_debounce(UInt32 delay_in_ticks,UInt32 keymask)
{
while(keybmp&keymask)
  {
  if(check_events())
    return(1);
  SysTaskDelay(delay_in_ticks);
  }

return(0);
}


void main_loop(void)
{
UInt32 ticks,ticks_last_update=0;
UInt32 tickdiff,delay_in_ticks=SysTicksPerSecond()/20;
int initpause;

InitHyperoid();
RestartHyperoid();

initpause=restore_vars();

while(check_events()==0)
  {
  if(overlaid_pause)
    {
    SysTaskDelay(delay_in_ticks);
    continue;
    }

  if(keybmp&keyBitHard4)
    {
    if(event_loop_debounce(delay_in_ticks,keyBitHard4))
      break;	/* exit while() */
    paused=!paused;
    }

  draw_select_buffer();
  
  if(paused)
    {
    if((keybmp&keyBitHard2) && (keybmp&keyBitHard3))
      {
      if(event_loop_debounce(delay_in_ticks,keyBitHard2|keyBitHard3))
        break;	/* exit while() */
      paused=0;
      NewGame();
      draw_deselect_buffer();
      continue;
      }
    
    SysTaskDelay(delay_in_ticks);
    drawpause();
    }
  else
    {
    /* not paused, so do proper delay. */
    if(prefs.battery_friendly)
      {
      int delaylen;
      
      /* hybrid approach - use SysTaskDelay for all but one
       * of the ticks, falling back on busy-wait for the last.
       */
      ticks=TimGetTicks();
      tickdiff=ticks-ticks_last_update;
      if(tickdiff<delay_in_ticks)
        {
        delaylen=delay_in_ticks-tickdiff;
        if(delaylen>1)
          SysTaskDelay(delaylen-1);
        }
      }
    
    /* traditional busy-wait */
    do
      {
      ticks=TimGetTicks();
      tickdiff=ticks-ticks_last_update;
      }
    while(tickdiff<delay_in_ticks);
    ticks_last_update=ticks;
    
    draw_clear_buffer();
    CheckScore();
    DrawObjects();
    }

  draw_copy_to_screen();
  
  if(paused)
    continue;
  
  if(bRestart)
    {
    restart_timer_count--;
    if(restart_timer_count==0)
      {
      bRestart = FALSE;
      RestartHyperoid();
      }
    }

  if(new_game)
    {
    new_game=0;
    NewGame();
    }

  /* initpause allows us to do one frame before pausing */
  if(initpause)
    {
    initpause=0;
    repause=0;	/* forget any pending one of these */
    paused=1;
    }
  
  /* if we've been told to do a frame before repausing by
   * main.c, the frame has already happened above. So now,
   * just pause.
   * redrawing when colour and/or invert state is changed
   * in the prefs.
   */
  if(repause)
    {
    repause=0;
    paused=1;
    continue;	/* don't toggle safe_draw */
    }
  
  safe_draw=!safe_draw;
  }

save_vars();
}
