/* Phyperoid 1.1 - an enhanced shoot-the-asteroids game.
 * Copyright (C) 1991 Edward Hutchins.
 * PalmOS changes copyright (C) 2002 Russell Marks.
 *
 * main.c - PilotMain() and PalmOS event-handling.
 */

#include <PalmOS.h>

#include "draw.h"
#include "misc.h"
#include "hyperoid.h"
#include "main-res.h"
#include "main.h"


UInt32 keybmp;
UInt32 keymask_orig;
int overlaid_pause=0,new_game=0,repause=0;


void keymask_init(void)
{
keymask_orig=KeySetMask(~(keyBitPageUp|keyBitPageDown|
                          keyBitHard1|keyBitHard2|
                          keyBitHard3|keyBitHard4));
}


void keymask_uninit(void)
{
KeySetMask(keymask_orig);
}


Err start_app(void)
{
UInt32 romver;

FtrGet(sysFtrCreator,sysFtrNumROMVersion,&romver);

/* must have OS >=2.0 */
if(romver<0x02003000)
  {
  FrmAlert(id_alert_romver);
  
  /* OS 1.0 will keep re-running us if we exit (!), so try
   * the launcher as a subtle fsck-off hint. ;-)
   */
  SysAppLauncherDialog();
  return(1);
  }

if(!draw_init())
  {
  FrmAlert(id_alert_probnomem);
  return(1);
  }

keymask_init();

FrmGotoForm(id_main_form);
check_events();

return(0);
}


void stop_app(void)
{
keymask_uninit();

draw_uninit();

FrmCloseAllForms();
}


#define get_obj_ptr(frm,id) \
	FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,id))


void do_prefs_form(void)
{
UInt16 button_id;
FormPtr frm;
ListPtr fg_list,bg_list;
int cols_changed=0;

frm=FrmInitForm(id_prefs_form);

/* quick sanity check just in case */
if(prefs.fg<0 || prefs.fg>7) prefs.fg=0;
if(prefs.bg<0 || prefs.bg>7) prefs.bg=7;

fg_list=get_obj_ptr(frm,id_prefs_fg_list);
bg_list=get_obj_ptr(frm,id_prefs_bg_list);

/* set up controls from prefs struct */
CtlSetValue(get_obj_ptr(frm,id_prefs_bf_check),prefs.battery_friendly);
CtlSetValue(get_obj_ptr(frm,id_prefs_invert_check),prefs.invert);
LstSetSelection(fg_list,prefs.fg);
LstSetSelection(bg_list,prefs.bg);
/* I don't actually need this with the size I've got them now, but... */
LstMakeItemVisible(fg_list,prefs.fg);
LstMakeItemVisible(bg_list,prefs.bg);
CtlSetLabel(get_obj_ptr(frm,id_prefs_fg_trigger),
            LstGetSelectionText(fg_list,prefs.fg));
CtlSetLabel(get_obj_ptr(frm,id_prefs_bg_trigger),
            LstGetSelectionText(bg_list,prefs.bg));

/* back to normal hard keys - mainly to allow up/down to be used
 * for the help string.
 */
keymask_uninit();

button_id=FrmDoDialog(frm);

keymask_init();

/* copy from controls back to prefs struct */
if(button_id==id_prefs_ok_button)
  {
  int invert,fg,bg;
  
  prefs.battery_friendly=CtlGetValue(get_obj_ptr(frm,id_prefs_bf_check));
  
  invert=CtlGetValue(get_obj_ptr(frm,id_prefs_invert_check));
  fg=LstGetSelection(fg_list);
  bg=LstGetSelection(bg_list);
  if(invert!=prefs.invert || fg!=prefs.fg || bg!=prefs.bg)
    cols_changed=1;
  prefs.invert=invert;
  prefs.fg=fg;
  prefs.bg=bg;
  }

FrmDeleteForm(frm);

/* if colour and/or invert state has changed and we're paused,
 * we need to run for a frame before repausing otherwise we
 * won't reflect the change. This is a bit unpleasant, but the
 * game code doesn't really support the notion of a redraw
 * without moving stuff, so...
 */
if(cols_changed && paused)
  paused=0,repause=1;
}


Boolean main_form_handle_event(EventType *event)
{
Boolean handled=false;

switch(event->eType)
  {
  case frmOpenEvent:
    FrmDrawForm(FrmGetActiveForm());
    handled=true;
    break;

  case menuEvent:
    switch(event->data.menu.itemID)
      {
      case id_main_menu_game_new:
        new_game=1;
        handled=true;
        break;
      case id_main_menu_game_pause:
        paused=!paused;
        handled=true;
        break;
      case id_main_menu_game_prefs:
        /* I think it's theoretically possible for this to
         * happen before prefs struct is initialised, so just
         * a little precaution against that...
         */
        if(prefs_ok)
          do_prefs_form();
        handled=true;
        break;
      case id_main_menu_help_instructions:
        keymask_uninit();
        FrmHelp(id_string_instructions);
        keymask_init();
        handled=true;
        break;
      case id_main_menu_help_keys:
        keymask_uninit();
        FrmHelp(id_string_keys);
        keymask_init();
        handled=true;
        break;
      case id_main_menu_help_about:
        FrmAlert(id_alert_help_about);
        handled=true;
        break;
      }
    break;
  
  default:
    break;
  }

return(handled);
}


Boolean app_handle_event(EventType *event)
{
FormPtr frm;
UInt16 form_id;
Boolean handled=false;

if(event->eType==frmLoadEvent)
  {
  /* init/activate form */
  form_id=event->data.frmLoad.formID;
  frm=FrmInitForm(form_id);
  FrmSetActiveForm(frm);

  /* set event handler for form */
  switch(form_id)
    {
    case id_main_form:
      FrmSetEventHandler(frm,main_form_handle_event);
      break;
    }

  handled=true;
  }

return(handled);
}


/* return non-zero if we want to exit */
int check_events(void)
{
EventType event;
UInt16 error;

/* poll the hard keys */
keybmp=KeyCurrentState();

do
  {
  EvtGetEvent(&event,0);
  if(event.eType==nilEvent) break;

  /* check for covered window */
  if(event.eType==winExitEvent)
    {
    if(event.data.winExit.exitWindow==(WinHandle)FrmGetFormPtr(id_main_form))
      overlaid_pause=1;
    }
  
  /* check for uncovered window :-) */
  if(event.eType==winEnterEvent)
    {
    if(event.data.winEnter.enterWindow==(WinHandle)FrmGetFormPtr(id_main_form) &&
       event.data.winEnter.enterWindow==(WinHandle)FrmGetFirstForm())
      overlaid_pause=0;
    }
  
  if(SysHandleEvent(&event))
    continue;
  
  if(MenuHandleEvent(NULL,&event,&error))
    continue;
  
  if(app_handle_event(&event))
    continue;

  FrmDispatchEvent(&event);
  }
while(event.eType!=appStopEvent);

return(event.eType==appStopEvent);
}


UInt32 PilotMain(UInt16 cmd,MemPtr cmdPBP,UInt16 launchFlags)
{
Err ret=0;		
	
if(cmd==sysAppLaunchCmdNormalLaunch)
  {
  if((ret=start_app())==0)
    {
    main_loop();
    stop_app();
    }
  }

return(ret);
}
